/* dis9830: a portable disassembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

#include <stdio.h>
#include <string.h>
#include "dis9830.h"
#include "opcode.h"

extern char *opbits[];

static int symcount = 0;

/* after having previously identified the OPCODE, decode the details of the instruction */

Inst *
decode(int opcode, unsigned short v, unsigned short address)
{
	char *p;
	int i = 0;
	char symname[32];
	Symbol *s;
	Inst *inst;

	if ((inst = (Inst *)malloc(sizeof(Inst))) == NULL)
		return(NULL);

	inst->v = v;	
	inst->address = address;
	inst->opcode = opcode;

	if (opcode == OPCODE_DATA) {
		inst->n = v;
		inst->flag = INST_DATA;
		return(inst);
	}

	for (p = opbits[opcode]; *p != 0; p++) {
		switch(*p) {
		case (' '):
			break;

		case ('1'):
		case ('0'):
		case ('.'):
			i++;
			break;

		case ('b'):
			inst->flag = INST_BRANCH;
			inst->n = (inst->n << 1) | ((v >> (15 - i++)) & 1);
			break;
		case ('m'):
			inst->flag = INST_MEM;
			inst->n = (inst->n << 1) | ((v >> (15 - i++)) & 1);
			break;
		case ('d'):
			inst->flag = INST_DEV;
			inst->n = (inst->n << 1) | ((v >> (15 - i++)) & 1);
			break;
		case ('n'):
			inst->flag = INST_SHIFT;
			inst->n = (inst->n << 1) | ((v >> (15 - i++)) & 1);
			break;

		case ('r'):
			inst->r = ((v >> (15 - i++)) & 1) ? 'B' : 'A';
			break;
		
		case ('g'):
			inst->g = ((v >> (15 - i++)) & 1) ? 'B' : 'A';
			break;

		case ('c'):
			if ((v >> (15 - i++)) & 1) 
				inst->mod |= MOD_CLEAR;
			break;
		case ('s'):
			if ((v >> (15 - i++)) & 1) 
				inst->mod |= MOD_SET;
			break;
		case ('i'):
			if ((v >> (15 - i++)) & 1) 
				inst->mod |= MOD_INDIRECT;
			break;
		case ('p'):
			if ((v >> (15 - i++)) & 1)
				inst->mod |= MOD_CURRENT;
			break;

		default:
			break;
		}
	}

	if (inst->flag & INST_MEM) {
		if (inst->mod & MOD_CURRENT) {
			inst->n |= (PAGEINDEX(address) << 11);
		}
	}

	if (inst->flag & INST_BRANCH) {
		if (inst->n & 0x10) {
			inst->n ^= 0x10;
			inst->n = address - inst->n;
		} else {
			inst->n = address + inst->n; 
		}
	}

	if (inst->flag & INST_SHIFT) {
		if ((opcode == OPCODE_SRR) || (opcode == OPCODE_RRR)) {
			inst->n += 1;
		}

		if ((opcode == OPCODE_ARR) || (opcode == OPCODE_SRL)) {
			inst->n = 16 - inst->n;
		}
	}
		
	if (inst->flag & (INST_MEM | INST_BRANCH)) {
		if ((s = findsymbol(inst->n)) == NULL) {
			sprintf(symname, "SYM_%d", symcount++);	
			inst->s = addsymbol(symname, inst->n);
		} else {
			inst->s = s;
		}
		inst->flag |= INST_SYM;

		/* 
		 * Heuristic: if a non-jump memory reference is direct, 
		 * then disassemble the address for the referenced symbol as data, 
		 * not an instruction 
		 *
		 * Since it's only a single pass, this strategy can cause cascading mis-idenfications; 
		 * doing better requires multiple passes and more bookkeeping. Probably not worth it.
		 *
 		 * The heuristic will not override explicit identifications provided
		 * on the command line with -C or -D.
		 */

		if ((opcode != OPCODE_JMP) && (opcode != OPCODE_JSM) && 
		     (inst->flag & INST_MEM) && !(inst->mod & MOD_INDIRECT)) {

			if (!(inst->s->flag & (SYM_FORCE | SYM_DATA))) {
				inst->s->flag = SYM_DATA;
				inst->s->reference = inst->address;
			}
		}
	} 

	/* 
	 * The assembler requires these instructions to explicitly 
	 * include the ,I modifier.
	 */

	if ((inst->opcode == OPCODE_STRg) || (inst->opcode == OPCODE_ISZg) || 
		(inst->opcode == OPCODE_DSZg) || (inst->opcode == OPCODE_JSMg)) {
			inst->mod |= MOD_INDIRECT;
	}

	return(inst);
}
