/* dis9830: a portable disassembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

#include <stdio.h>
#include <string.h>
#include "opcode.h"
#include "opasm.h"
#include "dis9830.h"

static Symbol *_slist = NULL;
static Inst *_iqueue = NULL;
static Inst *_iend;


/* 
 * destructively insert unsorted list input into sorted list output, 
 * using comparison function compare(); return sorted list. 
 */

typedef int (*CompareFunc)(Symbol *a, Symbol *b);

Symbol *
insertlist(Symbol *input, Symbol *output, CompareFunc compare)
{
	Symbol *tin, *tout,  *prev, *next, *sorted;

	tin = input;
	sorted = output;
	prev = NULL;

	while (tin) {
		next = tin->next;
		if (sorted == NULL) {
			sorted = tin;
			sorted->next = NULL;
		} else {
			prev = NULL;
			for (tout = sorted; tout != NULL; tout = tout->next) {
				if (compare(tin, tout) <= 0) {
					if (prev) {
						prev->next = tin;
						tin->next = tout;
					} else {
						tin->next = sorted;
						sorted = tin;
					}
					break;
				} else {
					prev = tout;
				}
			}
			if (tout == NULL) {
				prev->next = tin;
				tin->next = NULL;
			}
		}
		tin = next;
	}
	return(sorted);
}

/* comparison functions; compare by address and by name */

int
sortaddr(Symbol *a, Symbol *b)
{
	return(a->address - b->address);
}

int
sortname(Symbol *a, Symbol *b)
{
	return(strcmp(a->name, b->name));
}

/* manage our simplistic symbol table; just a linked list, sorted by address */

Symbol *
addsymbol(char *name, unsigned short address)
{
	Symbol *s; 

	if ((s = findsymbol(address)) != NULL) {
		return(s);
	}

	if ((s = (Symbol *)malloc(sizeof(Symbol))) == NULL)
		return(NULL);

	s->name = strdup(name);
	s->address = address;
	s->flag = 0;
	s->reference = 0;
	s->next = NULL;

	_slist = insertlist(s, _slist, sortaddr);

	return(s);
}

Symbol *
findsymbol(unsigned short address)
{
	Symbol *t;

	for (t = _slist; t != NULL; t = t->next) {
		if (t->address == address)
			return(t);

		if (t->address > address)
			return(NULL);
	}
	return(NULL);
}


/* dump the symbol table, sorted by name */

void
dumpsymbol(void)
{
	Symbol *s, *sorted = NULL;

	sorted = insertlist(_slist, sorted, sortname);

	for (s = sorted; s != NULL; s = s->next) {
		printf("%-32s: 0%06o", s->name, s->address);
		if (s->flag & SYM_DATA) {
			printf("; data"); 
		} else {
			printf("; instruction");
		}
		if (s->flag & SYM_FORCE) {
			printf(" (forced)");
		} else {
			if (s->flag & SYM_DATA) {
				printf(" (reference @ 0%06ho)", s->reference);
			}
		}
		printf("\n");
	}
}


/* maintain the list of decoded instructions so we can add the appropriate symbols as we discover them */

void
queueinst(Inst *i)
{
	if (_iqueue == NULL) {
		_iqueue = i;
		_iend = i;
	} else {
		_iend->next = i;
		_iend = i;
	}
}

/* find a particular previously queued instruciton by address */

Inst *
findinst(int address)
{
	Inst *i;

	for (i = _iqueue; i != NULL; i = i->next) {
		if (i->address == address)
			return(i);
	}
	return(NULL);
}



/* dump the decoded instruction, with symbols and other decorations as needed */

void 
dumpinst(int verbose)
{
	Inst *i;
	Symbol *s = _slist;
	char addr[32] = "";

	if (_iqueue)
		printf("%-24s          ORG 0%06ho\n\n", addr, _iqueue->address);

	while ((s != NULL) && (s->address < _iqueue->address)) {
		printf("%-24s%-8s  EQU 0%06ho\n", addr, s->name, s->address);
		s = s->next;
	}

	printf("\n");

	for (i = _iqueue; i != NULL; i = i->next) {

		if (verbose) {
			sprintf(addr, "0%06ho: 0%06ho  ", i->address, i->v);
		}

		if ((s != NULL) && (s->address == i->address)) {
			printf("%-24s%-8s  ", addr, s->name);
			if (s->flag & SYM_DATA) {
				/* code treats this address as data; disassmble it as such */
				printf("%s 0%06ho\n", opasm[OPCODE_DATA], i->v);
				s = s->next;
				continue;
			}
			s = s->next;
		} else {
			printf("%-24s          ", addr);
		}


		/* print the opcode (with embedded register, if appropriate) */

		if (i->r != 0) {
			printf(opasm[i->opcode], i->r);
		} else {
			printf("%s", opasm[i->opcode]);
		}

		if (i->g != 0) {
			printf(" %c", i->g);
		}

		if (i->flag & INST_SYM) {
			printf(" %s", i->s->name);
		} else {
			if (i->flag & (INST_DEV | INST_SHIFT | INST_DATA)) {
				printf(" 0%ho", i->n);
			}
		}

		if (i->mod & (MOD_CLEAR | MOD_SET | MOD_INDIRECT)) {
			printf(",");

			if (i->mod & MOD_CLEAR) 
				printf("C");

			if (i->mod & MOD_SET)
				printf("S");

			if (i->mod & MOD_INDIRECT)
				printf("I");
		}

		printf("\n");
	}
	
	printf("\n");

	while (s != NULL) {
		printf("%-24s%-8s  EQU 0%06ho\n", "", s->name, s->address);
		s = s->next;
	}
}
