/* dis9830: a portable disassembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include "opcode.h"
#include "opbits.h"
#include "header.h"
#include "dis9830.h"

static void
_usage(char *name)
{
	fprintf(stderr, "usage: %s [-x] [-v] [-o origin] [(-C|-D) <name>=<value>] [-f argfile] [-a outfile] ([-i] | [infile ...])\n", name);
}

// read from the input stream in the appropriate format; returns number of items read

static int
_getword(FILE *fp, int xns, int *addrp, unsigned short *dp)
{

	int c;
	int n;
	unsigned short num = 0;
	
	if (!xns) {
		return(fread(dp, 1, sizeof(short), fp));
	} else {
		while ((c = fgetc(fp)) != EOF) {
			n = -1;
			if (isxdigit(c)) {
				if (isalpha(c)) {
					n = toupper(c) - 'A' + 10;
				} else {
					n = c - '0';
				}
			}

			if (n >= 0) {
				num = (num << 4) | n;
			} else if (c == '~') { // number delimiter
				*dp = num;
				return(1);
			} else if (c == ':') { // address delimiter
				*addrp = num;
				num = 0;
			} else if (c == '!') { // EOF
				break;
			} else {
				num = 0;
			}
		}

		return(0);
	}
}

// predefine symbols/locations as code or data 
// 	s is a string of the form <name>=<value>; name can be null

static void
_predefine(char *s, int isdata)
{
	char *p;
	int address;
	Symbol *sym;

	if ((p = strchr(s, '=')) == NULL) {
		error("argument must match <label>=<value>");
		return;
	}

	*p++ = '\0';

	if (sscanf(p, "%i", &address) != 1) {
		error("symbol address not recognized: %s", p);
		return;
	}

	if ((sym = addsymbol(s, address)) == NULL) {
		error("couldn't add symbol %s, address 0%06o\n", &(s[2]), address);
		return;
	}

	// mark type as "forced", to prevent heuristic from changing meaning

	sym->flag = SYM_FORCE;
	if (isdata) {
		sym->flag |= SYM_DATA;
	}
}


int
main (int argc, char *argv[])
{
	char *pgmname = argv[0];
	int usestdin = 0;
	int verbose = 0;
	int address = 0;
	int format = FMT_HEADER;
	int i, j;
	unsigned short sum, count, data;
	int c, m;
	Node *t = NULL;
	FILE *infile;
	FILE *argfile = NULL;;
	Inst *inst = NULL;
	char *p;
	char buf[BUFSIZ];
	HP9830Header header;

	while ((c = getopt(argc, argv, "C:D:f:hxvio:")) != -1) {
		switch (c) {
		case('v'):
			verbose = 1;
			break;
		case('i'):
			usestdin = 1;
			break;
		case('x'):
			format |= FMT_XNS;
			break;
		case('f'):
			if (((argfile = fopen(optarg, "r")) == 0)) {
				error("Can't open argument file %s\n", optarg);
				return(-1);
			} 
			break;
		case('C'):
		case('D'):
			_predefine(optarg, (c == 'D'));
			break;
		case('o'):
			if (sscanf(optarg, "%i", &address) != 1) {
				error("origin address not recognized: %s", optarg);
				return(-1);
			}
			break;
		case('h'):
			format &= ~FMT_HEADER;
			break;
		case('?'):
		default:
			_usage(pgmname);
			return(-1);
		}
	}

	argc -= optind;
	argv += optind;

	if (usestdin && (argc > 0)) {
		error("non-option arguments follow -i\n");
		_usage(pgmname);
		return(-1);
	}

	if (!usestdin && (argc == 0)) {
		error("missing input filename(s), and -i not specified");
		_usage(pgmname);
		return(-1);
	}

	if (argfile) {
		while(fgets(buf, BUFSIZ, argfile) != NULL) {
			p = strchr(buf, '-');
			switch(p[1]) {
			case('C'):
			case('D'):
				i = strspn(&(p[2]), " \t");
				_predefine(&(p[i+2]), (p[1] == 'D'));
				break;
			default:
				error("argument file syntax error; argument not found: %s", buf);
				return(-1);
			}
		}
	}

	for (i = 0; i < (sizeof(opbits)/sizeof(char *)); i++) {
		t = insert(t, opbits[i], 0, 0, i);
	}

	for (i = 0; i < argc; i++) {

		if (((infile = fopen(argv[i], "r")) == 0)) {
			error("Can't open input file %s", argv[i]);
			return(-1);
		} 

		if (verbose) {
			printf("[Object file '%s':]\n\n", argv[i]);
		}

		/* deal with the header, if it exists. */

		if (format & FMT_HEADER) {

			for (j = 0; j < (sizeof(header)/sizeof(unsigned short)); j++) {
				if (_getword(infile, format & FMT_XNS, &address, &(header.data[j])) == 0) {
					error("failed to read header, offset %d", j);
					return(-1);
				}
			}

			if (header.magic != HP9830_MAGIC) {
				error("missing magic number in object header: found 0x%x", header.magic);
				return(-1);
			}

			if (header.ftype != HP9830_FT_BINARY) {
				error("filetype is not binary; found 0x%x", header.ftype);
				return(-1);
			}
		}

		count = 0;
		sum = 0;

		while (_getword(infile, format & FMT_XNS, &address, &data) != 0) {

			count++;
			if (format & FMT_HEADER) {
				if (count == (header.cursz + 1)) {
					if (sum != data) {
						error("object checksum does not match: calc: 0%06o, found: 0%06o", sum, data);
						return(-1);
					}
					break;
				} else {
					sum += data;
				}
			} 

			if ((m = match(t, data)) == -1) {
				m = OPCODE_DATA;
			} 
			if ((inst = decode(m, data, address)) != NULL)
				queueinst(inst);

			address++;
		}

		if ((format & FMT_HEADER) & feof(infile)) {
			error("checksum not found; logical EOF beyond end of input");
			return(-1);
		}

		if (_getword(infile, format & FMT_XNS, &address, &data) != 0) {
			error("data past logical EOF of input; ignored");
		}
		
		fclose(infile);
	}

	dumpinst(verbose);

	if (verbose) {
		printf("\n[Symbol table:]\n\n");
		dumpsymbol();
	}
		
	return(0);
}
