/* dis9830: a portable disassembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

#include "dis9830.h"

#ifdef DEBUG
void 
dumptree(Node *t)
{
	if (t == NULL)
		return;

	if (t->n == -1) {
		printf("node (terminal %p): opcode = %d\n", t, t->opcode);
	} else {
		printf("node (%p): n = %d, left = %p, right = %p\n", t, t->n, t->child[0], t->child[1]);
		dumptree(t->child[0]);
		dumptree(t->child[1]);
	}
}
#endif

/* duplicate a subtree; helper function, see below */

static Node *
_duplicate(Node *t)
{
	Node *d;

	if (t == NULL)
		return(NULL);

	if ((d = (Node *)malloc(sizeof(Node))) == NULL)
		return(NULL);

	d->n = t->n;
	if (d->n == -1) {
		d->opcode = t->opcode;
	} else {
		d->child[0] = _duplicate(t->child[0]);
		d->child[1] = _duplicate(t->child[1]);
	}
	return(d);
}


/* 
 * insert() builds a binary search tree which can be used for quickly matching 
 * opcode prototypes.
 *
 * The prototypes for the opcodes have bits that are "known" (i.e. explicitly 
 * specified as '1' or '0'), and other bits that are "don't care" (for the 
 * purpose of recognition)
 *
 * The tree branches at "known" bits, and elides "don't care" positions. 
 *
 * i is index into the prototype (bits[]), p is position in word, t is a pre-existing tree; 
 * returns the new tree after the insert.
 */

Node *
insert(Node *t, char *bits, int i, int p, int opcode)
{
	Node *t1;

	/* skip bits that are "don't care"; keep track of bit positions in both 
		opcode prototype string and in word */

	while (bits[i] != '0' && bits[i] != '1' && bits[i] != '\0') {
		if (bits[i] != ' ')
			p++;
		i++;
	}

	if (t == NULL) {
		if ((t = (Node *)malloc(sizeof(Node))) == NULL) {
			printf("panic\n");
			return(NULL);
		}
		t->n = p;
	}

	/* if we're at the end of the prototype, create terminal node */

	if (bits[i] == '\0') {
		t->n = -1;
		t->opcode = opcode;
		return(t);
	}

	/* 
	 * As we walk the tree, the node we encounter will either 
	 *
	 * 	- represent the next "known" bit position 
	 *	- represent a bit position that preceeds the next "known" bit; or
	 *	- represent a bit position that follows the next "known" bit
	 */

	if ((t->n == -1) || (t->n > p)) {

		/*
		 * either a terminal node, or one representing a bit that follows the one 
		 * we were looking for. In either case, we now care about a bit that was 
		 * previously "don't care", so add new intermediate state, duplicate the 
		 * existing subtree so that both child subtrees of the new node are the same, 
		 * and recursively insert remainder into the appropriate child subtree.
		 */

		t1 = t;	
		if ((t = (Node *)malloc(sizeof(Node))) == NULL) {
			printf("panic\n");
			return(NULL);
		}

		t->child[0] = t1;
		t->child[1] = _duplicate(t1);
		t->n = p;

		t->child[bits[i] - '0'] = insert(t->child[bits[i] - '0'], bits, i+1, p+1, opcode);
	} else if (t->n == p) {

		/* found a node that represents the bit we care about; insert rest into appropriate subtree */

		t->child[bits[i] - '0'] = insert(t->child[bits[i] - '0'], bits, i+1, p+1, opcode);
	} else {

		/* 
		 * By process of elmination, (t->n < p) 
	  	 *
		 * Node represents bit that preceeds the one we're looking for, and is "don't care" 
		 * for our purposes; therefore just continue insertion into both children.
		 */

		t->child[0] = insert(t->child[0], bits, i, p, opcode);
		t->child[1] = insert(t->child[1], bits, i, p, opcode);
	} 

	return(t);
}

/* 
 * Does v match a bit pattern in encoded in tree t? 
 * Returns the opcode (if found), -1 otherwise
 * (implicit instruciton length is 16 bits).
 */

int
match(Node *t, unsigned short v)
{
	Node *p = t;
	int i = 0;
	
	while (p != NULL) {

		if (p->n == -1) {
			return(p->opcode);
		}

		if (p->n == i) {
			p = p->child[(v >> (15 - i)) & 1];
		}

		i++;
	}

	return(-1);
}
